{ **************************************************************************** }
{ FileName............: FM1216.PAS                                             }
{ Project.............: BT8X8                                                  }
{ Author(s)...........: MM                                                     }
{ Version.............: 1.00                                                   }
{ ---------------------------------------------------------------------------- }
{ Definitions for tuner FM1216 (Philips) for CPH051 cards.                     }
{                                                                              }
{ Version  Date      Comment                                                   }
{   1.00   20010317  - Initial release                                         }
{ **************************************************************************** }
unit FM1216;


{ **************************************************************************** }
{                                                                              }
{ WRITE MODE                                                                   }
{ Data send to the FM1216 though the I2C is layout as follows:                 }
{  Byte  Definition                                                            }
{   0    Address (I2C), = $C0, $C2, $C4 or $C6                                 }
{        Note that this address is fixed through hardware                      }
{   1    Program divider byte 1 (MS byte, bit 7 always 0)                      }
{        BYTE = (  0 N14 N13 N12 N11 N10 N09 N08)                              }
{   2    Program divider byte 2 (LS byte)                                      }
{        BYTE = (N07 N06 N05 N04 N03 N02 N01 N00                               }
{        The program divider is calculated as follows:                         }
{          N = 16 * (Frf + Fif)    Frf = Picture frequency (MHz)               }
{                                  Fif = Crystal frequency (MHz) = hardware set}
{                                  * For PICTURE Fif = 38.9                    }
{                                    For COLOUR  Fif = 34.47                   }
{                                    For SOUND 1 Fif = 33.4                    }
{                                    For SOUND 2 Fif = 33.16                   }
{                                    For NICAM   Fif = 33.05                   }
{                                  * For FM      Fif = 10.7                    }
{          Fosc = N/16 (MHz)                                                   }
{        OR ALTERNATIVELY:                                                     }
{          Fosc = Frf + Fif                                                    }
{          N = 16 * Fosc                                                       }
{   3    Control information byte 1                                            }
{        BYTE = (  1  CP  T2  T1  T0 RSA RSB  OS)                              }
{        CP      = Charge Pump                                                 }
{             0  = moderate speed tuning                                       }
{             1  = fast tuning (recommended for TV, for FM only during         }
{                  actual scanning)                                            }
{        T2-T0   = Test mode settings                                          }
{           001  = normal operation                                            }
{        RSA-RSB = Ratio select bits                                           }
{            x0  = 50 kHz (FM band)                                            }
{            10  = 31.25 kHz (slow picture search)                             }
{            11  = 62.5  kHz (normal picture search)                           }
{        OS      = PLL disabling                                               }
{             0  = normal operation                                            }
{             1  = switches charge pump to high impedance state                }
{   4    Control information byte 2                                            }
{        BYTE = ( P7  P6  P5  P4  P3  P2  P1  P0)                              }
{        1010x1xx = FM: FM band    ( 87.50 - 108.0  MHz  +- 3.0 MHz)           }
{        xxxx0101 = FM: AFC (status can be read from A/D bits in status byte)  }
{                       Status bits A2-A0 indicate the following:              }
{                       A2-A0  000: -125 kHz                                   }
{                              001: -62.5 kHz                                  }
{                              010: 0                                          }
{                              011: +62.5 kHz                                  }
{                              100: +125 kHz                                   }
{        xxxx0100 = FM: RF  (level  can be read from A/D bits in status byte)  }
{        xxxx0110 = FM: Mono                                                   }
{        xxxx1100 = FM: Mute                                                   }
{        1010x0xx = TV: Low band   ( 48.25 - 168.25 MHz  +- 1.5 MHz)           }
{        1001x0xx = TV: Mid band   (175.25 - 447.25 MHz  +- 3.0 MHz)           }
{        0011x0xx = TV: High band  (455.25 - 855.25 MHz  +- 3.0 MHz)           }
{        xxxx0001 = TV: Power down mode                                        }
{        xxxx0000 = TV: System B, G negative mode                              }
{                                                                              }
{ Note: the <ProgramDivider> and <ControlWord> can be exchanged if desired     }
{       because the MS bit of the sent data idicates either a                  }
{       <ProgramDivider> (MSbit=0) or a <ControlWord> (MSbit=1).               }
{       Furthermore it is allowed to send only the first <Control> byte        }
{       if such is necessary, but only of it is the last byte send!            }
{                                                                              }
{ READ MODE                                                                    }
{ Data read from the FM1216 though the I2C is layout as follows:               }
{  Byte  Definition                                                            }
{   0    Address (I2C), = $C0, $C2, $C4 or $C6                                 }
{        Note that this address is fixed through hardware                      }
{   1    Status byte                                                           }
{        BYTE = (POR  FL  I2  I1  I0  A2  A1  A0)                              }
{        POR     = Power On Reset                                              }
{        FL      = In Lock flag                                                }
{             1  = Loop is phase-loacked (for at-least 8 periods)              }
{        I2-I0   = Digital information I/O ports P2-P0                         }
{        A2-A0   = A/D converter port P6                                       }
{                                                                              }
{                                                                              }
{ Because audio is first fed through the Bt8x8 chip, we must set the mute for  }
{ this chip too.                                                               }
{ **************************************************************************** }

interface

function Tuner_SetFM(Freq: integer; AFC: boolean): boolean;
function Tuner_SetFMFreq(Freq: integer): boolean;
function Tuner_SetTV(Freq: integer): boolean;
function Tuner_SetTVFreq(Freq: integer): boolean;
function Tuner_SetSound(Sound: boolean): boolean;
function Tuner_GetStatus: byte;


implementation

uses
  Windows,
  BTVID_32;

const
  CTuner_Settle    = 30;                                   { Time (ms) to settle PLL }
  CTuner_SettleLoop= 5;                                    { Max 150 ms for settling possible }
  CTuner_Address   = $C0;
  CTuner_TV_Fif    = 38900000;                             { IF (Hz) for TV }
  CTuner_FM_Fif    = 10700000;                             { IF (Hz) for FM }
  { Frequency ranges (MIN, [other, other,] MAX }
  CTuner_FM_Band  : array[0..1] of integer = (87500000, 108000000);
  CTuner_TV_Band  : array[0..3] of integer = (48250000, 170000000, 450000000, 855250000);

  { Control word }
  CTuner_Control              = $8000;                     { High bit always high }

  CTuner_Control_CP           = $4000;                     { Charge Pump bit }
  CTuner_Control_CP_Fast      = $4000;                     { Fast tuning mode }
  CTuner_Control_CP_Moderate  = $0000;                     { Moderate tuning mode }
  CTuner_Control_CP_TV        = $4000;                     { Recommended TV mode }
  CTuner_Control_CP_FM        = $0000;                     { Recommended FM mode }
  CTuner_Control_CP_FM_Search = $4000;                     { Recommended FM mode when searching }

  CTuner_Control_T012         = $3800;                     { Test mode settings }
  CTuner_Control_T012_Normal  = $0800;                     { Normal test mode settings }

  CTuner_Control_RSAB         = $0600;                     { Ratio select bits }
  CTuner_Control_RSAB_FM      = $0000;                     { Ratio select FM (50 kHz) }
  CTuner_Control_RSAB_TV_Slow = $0200;                     { Ratio select TV (31.25 kHz) slow picture search }
  CTuner_Control_RSAB_TV_Fast = $0600;                     { Ratio select TV (62.50 kHz) fast picture search }
  CTuner_Control_RSAB_Normal  = $0600;                     { Ratio select TV (62.50 kHz) normal }

  CTuner_Control_OS           = $0100;                     { PLL disabling bit }
  CTuner_Control_OS_Normal    = $0000;                     { PLL normal operation }
  CTuner_Control_OS_Disable   = $0100;                     { PLL disable (charge pump high impedance) }

  CTuner_Control_Band_FM      = $00A4;                     { FM band       , CTuner_FM_Band[0]..[1] }
  CTuner_Control_Band_TV_Low  = $00A0;                     { TV band (low) , CTuner_TV_Band[0]..[1] }
  CTuner_Control_Band_TV_Mid  = $0090;                     { TV band (mid) , CTuner_TV_Band[1]..[2] }
  CTuner_Control_Band_TV_High = $0030;                     { TV band (high), CTuner_TV_Band[2]..[3] }

  CTuner_Control_TV_PowerDown = $0001;                     { TV in power down }
  CTuner_Control_TV_BGNeg     = $0000;                     { TV system B/G negative mode }
  CTuner_Control_FM_AFC       = $0005;                     { FM, AFC in status }
  CTuner_Control_FM_RIF       = $0004;                     { FM, RF input level in status }
  CTuner_Control_FM_Mono      = $0006;                     { FM, Mono mode }
  CTuner_Control_FM_Mute      = $000C;                     { FM, Mute }

  { Status byte }
  CTuner_Status_POR      = $80;
  CTuner_Status_FL       = $40;
  CTuner_Status_FL_Lock  = $40;
  CTuner_Status_I012     = $18;

  { Status byte AFC (A2 A1 A0) }
  CTuner_Status_A012     = $07;
  CTuner_Status_AFC_M125 = $00;
  CTuner_Status_AFC_M62  = $01;
  CTuner_Status_AFC_0    = $02;
  CTuner_Status_AFC_P62  = $03;
  CTuner_Status_AFC_P125 = $04;


  { These depend on the TV card used }
  {  xxxxxx1111xxxxxxxx              }
  {        |||                       }
  {        || -----------  SOURCE 0  %00 = ?       %01 = EXT }
  {        |-------------  SOURCE 1  %10 = TUNER   %11 = ?   }
  {        --------------  MUTE      }
  GPIO_OUTEN     = $000F00;                                { The 4 bits used by the card    }

var
//  RetryI2C     : boolean;
  LastControlTV: word;

{ **************************************************************************** }
{ Params   : <Mute>    Mute or not                                             }
{ Return   : <Result>  TRUE for success                                        }
{                                                                              }
{ Descript : Select input source.                                              }
{ Notes    : The source is usually selected through the GPIO (General Purpose  }
{            I/O). The definition (use) of the GPIO can be different for TV    }
{            cards.                                                            }
{ **************************************************************************** }
function Tuner_SelectTuner: boolean;
var
  TheValue: integer;
  OrgValue: integer;
begin
  Result := false;
//  RetryI2C := true;

  OrgValue := GetGPOE;                                     { Get current setting }
  SetGPOE(OrgValue or GPIO_OUTEN);                         { Enable the output bits }
  TheValue := $02;                                         { SOURCE = %10 }
  if SetGPDataBits(9, 10, TheValue, 0)<>Success then Exit;
  SetGPOE(OrgValue);                                       { Restore original setting }
  Result := true;
end;


{ **************************************************************************** }
{ Params   : <Sound>   TRUE for sound                                          }
{                      FALSE to mute                                           }
{ Return   : <Result>  TRUE for success                                        }
{                                                                              }
{ Descript : Set sound status. Always sets input from tuner.                   }
{ Notes    : Sound is usually enabled through the GPIO (General Purpose I/O).  }
{            The definition (use) of the GPIO can be different for TV cards.   }
{ **************************************************************************** }
function Tuner_SetSound(Sound: boolean): boolean;
var
  TheValue: integer;
begin
  Result := false;
//  RetryI2C := true;

  SetGPOE(GetGPOE or GPIO_OUTEN);                          { Enable the output bits }
  if Sound then TheValue := $02                            { %010 }
           else TheValue := $06;                           { %110 }
  if SetGPDataBits(9, 11, TheValue, 0)<>Success then Exit;
  Result := true;
end;


{ **************************************************************************** }
{ Params   : <Address>  I2C Address                                            }
{            <Status>   Received status                                        }
{ Return   : <Result>   TRUE for success                                       }
{                                                                              }
{ Descript : Read tuner data from I2C (status).                                }
{ Notes    : -                                                                 }
{ **************************************************************************** }
function Tuner_ReadI2C(Address: byte; var Status: byte): boolean;
begin
  Result := false;
  if (I2CHWRead(Address, Status) = Fail) then Exit;
  Result := true;
end;


{ **************************************************************************** }
{ Params   : -                                                                 }
{ Return   : <Result>  Status byte                                             }
{                      POR  INLOCK  P2  P1  P0  A2  A1  A0                     }
{                                   \DIG INFO/  \   A/D  /                     }
{                                   0                        = Radio           }
{                                   1                        = TV              }
{                                       0                    = MONO   (?)      }
{                                       1                    = STEREO (?)      }
{                                                                              }
{ Descript : Get status from tuner.                                            }
{ Notes    : -                                                                 }
{ **************************************************************************** }
function Tuner_GetStatus: byte;
begin
  if not(Tuner_ReadI2C(CTuner_Address, Result)) then Result := 0;
end;


{ **************************************************************************** }
{ Params   : <Freq>    Frequency (Hz) to set frequency to                      }
{            <AFC>     TRUE  if AFC for status                                 }
{                      FALSE if RF  for status                                 }
{ Return   : <Result>  TRUE for success                                        }
{                                                                              }
{ Descript : Set FM (radio) mode                                               }
{ Notes    : -                                                                 }
{ **************************************************************************** }
function Tuner_SetFM(Freq: integer; AFC: boolean): boolean;
var
  Fosc        : real;
  ProgramWord : word;
  ControlWord : word;
  Loop        : byte;
  Control     : word;
begin
  Result := false;
  if Freq < CTuner_FM_Band[System.Low (CTuner_FM_Band)] then Exit;
  if Freq > CTuner_FM_Band[System.high(CTuner_FM_Band)] then Exit;

  Fosc := (CTuner_FM_Fif + Freq);                          { Oscillating frequency tuner }
  Fosc := 16 * Fosc;                                       { The divider needs a 16* value }
  Fosc := Fosc / 1000000;                                  { Convert to MHz }


  ProgramWord := round(Fosc) and $7FFF;                    { High bit ALWAYS low }
  I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  if RetryI2C then I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  RetryI2C := false;

  if (Tuner_GetStatus and $20)<>0 then
  begin
    { Before we set the FM radio mode we must set the TV band to LOW first. }
    { If we don't then we can not always set the FM band for some reason (too large a step?). }
    { Since we are anyway, we can also disable the TV tuner. }
    ControlWord := CTuner_Control or                         { Required high bit ALWAYS high }
                   CTuner_Control_CP_Moderate or             { Charge pump }
                   CTuner_Control_T012_Normal or             { Test mode }
                   CTuner_Control_RSAB_Normal or             { Ratio select }
                   CTuner_Control_OS_Normal or               { PLL }
                   CTuner_Control_Band_TV_Low or             { Band }
                   CTuner_Control_TV_PowerDown;              { Mode }
    I2CHWWrite3(CTuner_Address, System.Hi(ControlWord), System.Lo(ControlWord));
    Sleep(30);
  end;

  if AFC then Control := CTuner_Control_FM_AFC
         else Control := CTuner_Control_FM_RIF;
  ControlWord := CTuner_Control or                         { Required high bit ALWAYS high }
                 CTuner_Control_CP_FM or                   { Charge pump }
                 CTuner_Control_T012_Normal or             { Test mode }
                 CTuner_Control_RSAB_Normal or             { Ratio select MUST BE NORMAL... }
                 CTuner_Control_OS_Normal or               { PLL }
                 CTuner_Control_Band_FM or                 { Band }
                 Control;                                  { Mode }
  I2CHWWrite3(CTuner_Address, System.Hi(ControlWord), System.Lo(ControlWord));

  for Loop := 1 to CTuner_SettleLoop do
  begin
    Sleep(30);
    if (Tuner_GetStatus and $40)<>0 then Break;
  end;
  Sleep(1);
  Result := true;
end;


{ **************************************************************************** }
{ Params   : <Freq>    Frequency (Hz) to set frequency to                      }
{ Return   : <Result>  TRUE for success                                        }
{                                                                              }
{ Descript : Set FM (radio) frequency only.                                    }
{ Notes    : -                                                                 }
{ **************************************************************************** }
function Tuner_SetFMFreq(Freq: integer): boolean;
var
  Fosc        : real;
  ProgramWord : word;
  Loop        : byte;
begin
  Result := false;
  if Freq < CTuner_FM_Band[System.Low (CTuner_FM_Band)] then Exit;
  if Freq > CTuner_FM_Band[System.high(CTuner_FM_Band)] then Exit;

  Fosc := (CTuner_FM_Fif + Freq);                          { Oscillating frequency tuner }
  Fosc := 16 * Fosc;                                       { The divider needs a 16* value }
  Fosc := Fosc / 1000000;                                  { Convert to MHz }

  ProgramWord := round(Fosc) and $7FFF;                    { High bit ALWAYS low }
  I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  if RetryI2C then I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  RetryI2C := false;

  for Loop := 1 to CTuner_SettleLoop do
  begin
    Sleep(30);
    if (Tuner_GetStatus and $40)<>0 then Break;
  end;
  Sleep(1);
  Result := true;
end;


{ **************************************************************************** }
{ Params   : <Freq>    Frequency (Hz) to set frequency to                      }
{ Return   : <Result>  TRUE for success                                        }
{                                                                              }
{ Descript : Set TV mode.                                                      }
{ Notes    : -                                                                 }
{ **************************************************************************** }
function Tuner_SetTV(Freq: integer): boolean;
var
  Fosc        : real;
  ProgramWord : word;
  ControlWord : word;
  Band        : word;
  Loop        : byte;
begin
  Result := false;
  if Freq < CTuner_TV_Band[System.Low (CTuner_TV_Band)] then Exit;
  if Freq > CTuner_TV_Band[System.high(CTuner_TV_Band)] then Exit;

  Band := CTuner_Control_Band_TV_High;
  if Freq < CTuner_TV_Band[2] then Band := CTuner_Control_Band_TV_Mid;
  if Freq < CTuner_TV_Band[1] then Band := CTuner_Control_Band_TV_Low;

  Fosc := (CTuner_TV_Fif + Freq);                          { Oscillating frequency tuner }
  Fosc := 16 * Fosc;                                       { The divider needs a 16* value }
  Fosc := Fosc / 1000000;                                  { Convert to MHz }

  ProgramWord := round(Fosc) and $7FFF;                    { High bit ALWAYS low }
  I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  if RetryI2C then I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  RetryI2C := false;

  ControlWord := CTuner_Control or                         { Required high bit ALWAYS high }
                 CTuner_Control_CP_TV or                   { Charge pump }
                 CTuner_Control_T012_Normal or             { Test mode }
                 CTuner_Control_RSAB_Normal or             { Ratio select }
                 CTuner_Control_OS_Normal or               { PLL }
                 Band or                                   { Band }
                 CTuner_Control_TV_BGNeg;                  { Mode }

  I2CHWWrite3(CTuner_Address, System.Hi(ControlWord), System.Lo(ControlWord));
  LastControlTV := ControlWord;

  for Loop := 1 to CTuner_SettleLoop do
  begin
    Sleep(30);
    if (Tuner_GetStatus and $40)<>0 then Break;
  end;
  Sleep(1);
  Result := true;
end;


{ **************************************************************************** }
{ Params   : <Freq>    Frequency (Hz) to set frequency to                      }
{ Return   : <Result>  TRUE for success                                        }
{                                                                              }
{ Descript : Set TV frequency.                                                 }
{ Notes    : -                                                                 }
{ **************************************************************************** }
function Tuner_SetTVFreq(Freq: integer): boolean;
var
  Fosc        : real;
  ProgramWord : word;
  ControlWord : word;
  Band        : word;
  Loop        : byte;
begin
  Result := false;
  if Freq < CTuner_TV_Band[System.Low (CTuner_TV_Band)] then Exit;
  if Freq > CTuner_TV_Band[System.high(CTuner_TV_Band)] then Exit;

  Band := CTuner_Control_Band_TV_High;
  if Freq < CTuner_TV_Band[2] then Band := CTuner_Control_Band_TV_Mid;
  if Freq < CTuner_TV_Band[1] then Band := CTuner_Control_Band_TV_Low;

  Fosc := (CTuner_TV_Fif + Freq);                          { Oscillating frequency tuner }
  Fosc := 16 * Fosc;                                       { The divider needs a 16* value }
  Fosc := Fosc / 1000000;                                  { Convert to MHz }

  ProgramWord := round(Fosc) and $7FFF;                    { High bit ALWAYS low }
  I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  if RetryI2C then I2CHWWrite3(CTuner_Address, System.Hi(ProgramWord), System.Lo(ProgramWord));
//  RetryI2C := false;

  ControlWord := CTuner_Control or                         { Required high bit ALWAYS high }
                 CTuner_Control_CP_TV or                   { Charge pump }
                 CTuner_Control_T012_Normal or             { Test mode }
                 CTuner_Control_RSAB_Normal or             { Ratio select }
                 CTuner_Control_OS_Normal or               { PLL }
                 Band or                                   { Band }
                 CTuner_Control_TV_BGNeg;                  { Mode }

  if ControlWord<>LastControlTV
    then I2CHWWrite3(CTuner_Address, System.Hi(ControlWord), System.Lo(ControlWord));
  LastControlTV := ControlWord;

  for Loop := 1 to CTuner_SettleLoop do
  begin
    Sleep(30);
    if (Tuner_GetStatus and $40)<>0 then Break;
  end;
  Sleep(1);
  Result := true;
end;


initialization
//  RetryI2C := true;
  LastControlTV := 0;

finalization

end.
